use crate::CvfsResult;
use log::debug;
use multi_key_manager::{file::File, OpeningNonce};
use ring::aead::{Aad, NonceSequence, OpeningKey, SealingKey};
use std::io::{Read, Seek, Write};

/// This copies the buffer during encryption so the original buffer stays untouched
#[must_use]
pub(crate) fn encrypt_file_buffered<N: NonceSequence>(
    real_file: &mut File,
    buffer: &[u8],
    //keys: &mut dyn LiveKeyMaterial,
    //key: &mut SealingKey<BorrowedNonce>,
    key: &mut SealingKey<N>,
) -> CvfsResult<()> {
    let mut buf = Vec::with_capacity(4096);
    buf.extend(buffer);

    //let mut key = keys.get_sealing_master_key();
    let aad = Aad::empty();
    key.seal_in_place_append_tag(aad, &mut buf)?;

    real_file.rewind()?;
    real_file.set_len(buf.len() as u64)?;
    real_file.write_all(&buf)?;
    real_file.sync_all()?;

    Ok(())
}

#[must_use]
pub(crate) fn decrypt_file_buffered(
    real_file: &mut File,
    //keys: &mut dyn LiveKeyMaterial,
    key: &mut OpeningKey<OpeningNonce>,
) -> CvfsResult<Vec<u8>> {
    // 4KB is a good step size for disk reads
    let mut buf = Vec::with_capacity(4096);
    real_file.rewind()?;
    real_file.read_to_end(&mut buf)?;

    //let mut key = keys.get_opening_master_key();
    let aad = Aad::empty();
    debug!(" > decrypt_file_buffered: {:?}", buf.len());
    let decrypted = key.open_in_place(aad, &mut buf)?;
    let decrypted_len = decrypted.len();

    buf.truncate(decrypted_len);

    Ok(buf)
}

#[cfg(feature = "speed_test")]
use std::time::{Duration, Instant};
#[cfg(feature = "speed_test")]
pub(crate) fn encrypt_decrypt_in_memory_speedtest(
    keys: &mut dyn multi_key_manager::LiveKeyMaterial,
    plain_buffer: &[u8],
) -> CvfsResult<(Duration, Duration)> {
    let mut buf = Vec::with_capacity(4096);
    buf.extend(plain_buffer);
    let aad = Aad::empty();

    let encryption_start = Instant::now();
    keys.get_sealing_master_key()
        .seal_in_place_append_tag(aad, &mut buf)?;
    let encryption_elapsed = encryption_start.elapsed();

    let aad = Aad::empty();
    let decryption_start = Instant::now();
    keys.get_opening_master_key().open_in_place(aad, &mut buf)?;
    let decryption_elapsed = decryption_start.elapsed();

    Ok((encryption_elapsed, decryption_elapsed))
}
