#![forbid(unsafe_code)]

use multi_key_manager::{file::open_rwlocked, LiveKeyMaterial, LiveKeys, WithNonce};
use std::{
    fs::create_dir_all,
    io::Seek,
    path::{Path, PathBuf},
    sync::{Arc, Mutex, Weak},
    thread::panicking,
};

pub mod errors;
use errors::{specify_io_error, specify_kme_error, CvfsErrors};
mod aead;
use aead::{decrypt_file_buffered, encrypt_file_buffered};
pub mod metadata;
use metadata::{FileInfo, FileTag, Metadata};
mod secure_file;
use log::{debug, error, trace};
use secure_file::ActiveFile;
pub use {errors::CvfsResult, secure_file::SecureFileHandler};

const SUPERBLOCK_PATH: &str = "encfsstorage.superblock";
const BLOCK_PATH: &str = "blocks/";
const METADATA_PATH: &str = "metadata.index";

pub fn open_dir<P: AsRef<Path>>(
    decryption_input: &str,
    directory_path: P,
) -> Result<CryptoVFS, CvfsErrors> {
    let directory_path = directory_path.as_ref().to_owned();
    specify_io_error(create_dir_all(&directory_path), "unable to create_dir_all")?;
    let keys_path = path_extend(&directory_path, SUPERBLOCK_PATH);
    if keys_path.is_dir() {
        return Err(CvfsErrors::GeneralIO(format!(
            "keys_path must not be a directory: {:?}",
            keys_path.display()
        )));
    }
    let metadata_path = path_extend(&directory_path, METADATA_PATH);
    if metadata_path.is_dir() {
        return Err(CvfsErrors::GeneralIO(format!(
            "metadata_path must not be a directory: {:?}",
            metadata_path.display()
        )));
    }
    let mut metadata_file = specify_io_error(open_rwlocked(&metadata_path), {
        #[cfg(feature = "fs_locks")]
        {
            "unable to lock the metadata"
        }
        #[cfg(not(feature = "fs_locks"))]
        {
            "unable to open the metadata"
        }
    })?;

    let blocks_path = path_extend(&directory_path, BLOCK_PATH);
    specify_io_error(create_dir_all(blocks_path), "unable to create blocks_path")?;

    let keys = if keys_path.is_file() {
        LiveKeys::read_from_disk(decryption_input, keys_path)
    } else {
        LiveKeys::generate_new(decryption_input, keys_path)
    };
    let mut keys = specify_kme_error(keys, "unable to read or generate master key")?;

    let metadata = if metadata_file.empty()? {
        Metadata::empty()
    } else {
        let content =
            decrypt_file_buffered(&mut metadata_file, &mut keys.get_opening_master_key())?;
        trace!(" > CryptoVfsInner::open_dir");
        Metadata::restore_from_slice(&content)?
    };

    Ok(CryptoVFS(Arc::new(Mutex::new(CryptoVfsInner {
        keys,
        metadata: metadata,
        file_tag_counter: 0, // AtomicUsize::new(0),
        real_base_path: directory_path,
    }))))
}

/// A mounted virtual file system lock wrapper
#[derive(Debug)]
pub struct CryptoVFS(Arc<Mutex<CryptoVfsInner>>);
impl CryptoVFS {
    pub fn exists(&self, path: &str) -> CvfsResult<bool> {
        self.0.lock()?.exists(path)
    }

    /// Flush all open SecureFileHandler and the metadata to disk
    ///
    /// Droping the VFS will automatically flush it
    pub fn flush_all(&mut self) -> CvfsResult<()> {
        self.0.lock()?.flush_all()
    }

    /// Opens a file in write-only mode.
    ///
    /// This function will create a file if it does not exist, and will truncate it if it does.
    pub fn create(&mut self, path: &str) -> CvfsResult<SecureFileHandler> {
        let this = Arc::downgrade(&self.0);
        self.0.lock()?.create(this, path)
    }

    /// Existing file handlers will return a `CvfsErrors::FileHandlerNotFound` on any subsequent action
    ///
    /// Returns true if the File was removed with this action, false if it did not exist
    pub fn delete(&mut self, path: &str) -> CvfsResult<bool> {
        self.0.lock()?.delete(path)
    }
}

/// A mounted virtual file system
#[derive(Debug)]
pub struct CryptoVfsInner {
    /// Cyclic reference to itself
    //this: Weak<Mutex<CryptoVfsInner>>,
    keys: LiveKeys,
    metadata: Metadata,
    file_tag_counter: u64, //AtomicUsize,
    real_base_path: PathBuf,
}
impl CryptoVfsInner {
    /// Opens a file in write-only mode.
    ///
    /// This function will create a file if it does not exist, and will truncate it if it does.
    fn create(
        &mut self,
        this: Weak<Mutex<CryptoVfsInner>>,
        path: &str,
    ) -> CvfsResult<SecureFileHandler> {
        //let weak_metadata = Arc::downgrade(&self.metadata);
        //let mut metadata = self.metadata.lock()?;
        //Arc::get_mut(&mut self.metadata).expect("CryptoVfsInner::create double get_mut(metadata)");
        let Self {
            metadata,
            file_tag_counter,
            ..
        } = self;

        let entry: &mut FileInfo = metadata.files.entry(path.into()).or_insert_with(|| {
            let mut real_path = path_extend(&self.real_base_path, BLOCK_PATH);
            let mut real_block_name;
            loop {
                let num = random_u64();
                real_block_name = format!("{num:8x}.bin");
                real_path.push(&real_block_name);
                if real_path.exists() == false {
                    // this is very likely
                    break;
                } else {
                    // remove the file name and try again
                    real_path.pop();
                }
            }
            debug!("new block path: {:?}", real_path.display());
            FileInfo::new(real_block_name)
        });
        debug!("CryptoVfsInner::create found entry: {entry:?}");

        let id = match &mut entry.active {
            Some(active) => {
                // already open
                active.rewind()?;
                active.id.clone()
            }
            None => {
                // try to open the file
                let id = FileTag::next(file_tag_counter);
                let mut file_key = self
                    .keys
                    .get_opening_master_key_with_nonce(entry.get_opening_nonce());

                // actually open the backing storage
                let mut real_path = path_extend(&self.real_base_path, BLOCK_PATH);
                real_path.push(&entry.real_block_name);
                let active = ActiveFile::open_rwlocked(&real_path, id.clone(), &mut file_key)?;

                debug!("CryptoVfsInner::create made a file: {:?}", active);
                entry.active = Some(active);

                id
            }
        };

        Ok(SecureFileHandler {
            id,
            cvfs: this, //Arc::downgrade(&self.metadata),
        })
    }

    fn exists(&self, path: &str) -> CvfsResult<bool> {
        Ok(self.metadata.files.contains_key(path))
    }

    /// Flush all open SecureFileHandler and the metadata to disk
    ///
    /// Droping the VFS will automatically flush it
    fn flush_all(&mut self) -> CvfsResult<()> {
        self.keys.store_to_disk()?;

        let metadata_path = path_extend(&self.real_base_path, METADATA_PATH);
        let metadata_file = open_rwlocked(metadata_path)?;
        //let mut metadata = self.metadata.lock()?;
        self.metadata.store_to_disk(metadata_file, &mut self.keys)
    }

    fn delete(&mut self, path: &str) -> CvfsResult<bool> {
        let found = if let Some(file_info) = self.metadata.files.remove(path) {
            let mut block_path = path_extend(&self.real_base_path, BLOCK_PATH);
            block_path.push(&file_info.real_block_name);
            std::fs::remove_file(block_path)?;
            true
        } else {
            false
        };
        Ok(found)
    }
}
impl Drop for CryptoVfsInner {
    fn drop(&mut self) {
        if let Err(e) = self.flush_all() {
            let error = format!("aead_vfs::CryptoVfsInner was not flushed to disk: {e:?}");
            if panicking() {
                error!("{error}");
            } else {
                panic!("{error}");
            }
        }
    }
}

pub fn path_extend<B: AsRef<Path>, E: AsRef<Path>>(base: B, ext: E) -> PathBuf {
    let mut buf = base.as_ref().to_owned();
    buf.push(ext);
    buf
}

/// Generate a secure random value
pub fn random<T: ring::rand::RandomlyConstructable>() -> T {
    let rng = ring::rand::SystemRandom::new();
    ring::rand::generate(&rng)
        .expect("unable to generate random value")
        .expose()
}
pub fn random_u64() -> u64 {
    let r = random::<[u8; 8]>();
    u64::from_le_bytes(r)
}

#[cfg(feature = "speed_test")]
pub fn do_speed_test(n_mb: usize, cvfs: &mut CryptoVFS) -> CvfsResult<String> {
    use byte_slice_cast::AsByteSlice;
    use std::{
        io::Write,
        mem::size_of,
        time::{Duration, Instant},
    };

    fn run(
        n_mb: usize,
        cvfs: &mut CryptoVFS,
    ) -> CvfsResult<(Duration, Duration, Duration, Duration, Duration, Duration)> {
        /////////////////
        let type_size = size_of::<u64>();
        let allocation_start = Instant::now();
        let len = 1024 * 1024 * n_mb / type_size;
        let mut buffer = Vec::with_capacity(len);
        let mut seed = 8739845878359807234u64;
        for _i in 0..len {
            buffer.push(seed);
            seed = seed.rotate_left(13) + 23;
        }

        /////////////////
        let random_name = format!("{}.bin", random_u64());
        debug!("do_speed_test < delete {:?}", cvfs.delete(&random_name));
        let mut fd = cvfs.create(&random_name)?;
        let u8_buffer = buffer.as_byte_slice();
        let allocation_elapsed = allocation_start.elapsed();

        let ram_start = Instant::now();
        fd.write_all(u8_buffer)?;
        let ram_elapsed = ram_start.elapsed();
        debug!("do_speed_test < written");

        let drop_start = Instant::now();
        drop(fd);
        let drop_elapsed = drop_start.elapsed();

        /////////////////
        debug!("do_speed_test < cleanup");
        let deletion_start = Instant::now();
        cvfs.delete(&random_name)?;
        let deletion_elapsed = deletion_start.elapsed();

        let keys = &mut cvfs.0.lock()?.keys;
        let (encryption_elapsed, decryption_elapsed) =
            aead::encrypt_decrypt_in_memory_speedtest(keys, u8_buffer)?;

        Ok((
            allocation_elapsed,
            ram_elapsed,
            drop_elapsed,
            deletion_elapsed,
            encryption_elapsed,
            decryption_elapsed,
        ))
    }

    // calculate stats and output in milliseconds at nano second resolution
    fn stats<I: Iterator<Item = Duration> + Clone>(it: I) -> (f64, f64) {
        let mut avg = Duration::new(0, 0);
        let mut len = 0u32;
        for e in it.clone() {
            avg += e;
            len += 1;
        }
        avg /= len;

        let avg_nanos = avg.as_nanos();
        // nano seconds is accurate enough and the Duration type does not support square or square root
        let mut dev_nanos = 0;
        for e in it {
            let delta = e.as_nanos() - avg_nanos;
            dev_nanos += delta * delta;
        }
        dev_nanos /= len as u128;
        let dev = (dev_nanos as f64).sqrt();

        (avg_nanos as f64 / 1000_000.0, dev / 1000_000.0)
    }

    let runs: Vec<(Duration, Duration, Duration, Duration, Duration, Duration)> = (0..16)
        .map(|_| run(n_mb, cvfs))
        .collect::<CvfsResult<_>>()?;

    let (alloc_avg, alloc_dev) = stats(runs.iter().map(|d| d.0));
    let (ram_avg, ram_dev) = stats(runs.iter().map(|d| d.1));
    let (drop_avg, drop_dev) = stats(runs.iter().map(|d| d.2));
    let (del_avg, del_dev) = stats(runs.iter().map(|d| d.3));
    let (enc_avg, enc_dev) = stats(runs.iter().map(|d| d.4));
    let (dec_avg, dec_dev) = stats(runs.iter().map(|d| d.5));

    let runtime = format!(
        "{} {} ({} bit) -- CPU features {:?}",
        target::arch(),
        target::os(),
        target::pointer_width(),
        target::features()
    );

    // nummerically unstable:  ((n_mb as f64) / 1024.0) / ((enc_avg / 1000) as f64);
    // nummerically stable:
    let ram_gbps = (((1000 * n_mb) as f64) / 1024.0) / ram_avg;
    let enc_gbps = (((1000 * n_mb) as f64) / 1024.0) / enc_avg;
    let dec_gbps = (((1000 * n_mb) as f64) / 1024.0) / dec_avg;

    Ok(format!(
        r#" ::: CryptoVFS Speed Test :::
* test size: {n_mb}MB
* allocation time: {alloc_avg:.3}ms  ±{alloc_dev:.3}ms
* RAM transfer time: {ram_avg:.3}ms  ±{ram_dev:.3}ms {ram_gbps:.3}GB/s
* enc + to disk time: {drop_avg:.3}ms  ±{drop_dev:.3}ms
* deletion time: {del_avg:.3}ms  ±{del_dev:.3}ms
* encrypt time: {enc_avg:.3}ms  ±{enc_dev:.3}ms {enc_gbps:.3}GB/s
* decrypt time: {dec_avg:.3}ms  ±{dec_dev:.3}ms {dec_gbps:.3}GB/s
* runtime: {runtime}
"#
    ))
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::io::{read_to_string, Read, Write};

    /// ext must not start with a `/`
    fn path<P: AsRef<Path>>(ext: P) -> PathBuf {
        let mut p = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        // we are in a cargo workspace
        p.pop();
        p.push("target");
        p.push("aead-vfs_tests");
        std::fs::create_dir_all(&p).expect("unable to create testing directory");
        p.push(ext);
        p
    }

    #[test]
    fn it_works() {
        let result = open_dir("bla bla secure key bla", path("it_works"));
        assert!(result.is_ok());
    }

    #[test]
    fn encrypt_decrypt_cycle() {
        let mut keys =
            LiveKeys::generate_new("paper_key paper_key", path("encrypt_decrypt_cycle.keys"))
                .unwrap();
        let mut file = open_rwlocked(path("encrypt_decrypt_cycle.bin")).unwrap();
        let random_data = b"abcdefghijklmnopqrstuvwxyz";
        println!(" >> DBG: {}", random_data.len());

        println!("nonce_before_encrypt: {:?}", keys.get_opening_nonce());
        encrypt_file_buffered(&mut file, random_data, &mut keys.get_sealing_master_key())
            .expect("unable to encrypt");

        let nonce_after_encrypt = keys.get_opening_nonce();
        println!("nonce_after_encrypt: {nonce_after_encrypt:?}");

        let decrypted = decrypt_file_buffered(&mut file, &mut keys.get_opening_master_key())
            .expect("unable to decrypt");

        let nonce_after_decrypt = keys.get_opening_nonce();
        println!("nonce_after_decrypt: {nonce_after_decrypt:?}");

        assert_eq!(random_data, &*decrypted);
    }

    /*
        #[test]
        fn encrypt_decrypt_cycle_large_file() {
            let mut keys = LiveKeys::generate_new(
                "paper_key paper_key",
                path("encrypt_decrypt_cycle_large_file.keys"),
            )
            .unwrap();
            let mut file = open_rwlocked(path("encrypt_decrypt_cycle_large_file.bin")).unwrap();
            let base_date = b"abcdefghijklmnopqrstuvwxyz";
            let mut random_data = Vec::new();
            for _ in 0..1024 * 1024 {
                random_data.extend_from_slice(base_date);
            }

            println!(" >> DBG: {}", random_data.len());

            println!("nonce_before_encrypt: {:?}", keys.get_opening_nonce());
            encrypt_file_buffered(&mut file, &random_data, &mut keys).expect("unable to encrypt");

            let nonce_after_encrypt = keys.get_opening_nonce();
            println!("nonce_after_encrypt: {nonce_after_encrypt:?}");

            let decrypted = decrypt_file_buffered(&mut file, &mut keys).expect("unable to decrypt");

            let nonce_after_decrypt = keys.get_opening_nonce();

            assert_eq!(random_data, decrypted);
        }
    // */

    #[test]
    fn double_store_and_decrypt_cycle() {
        let mut keys = LiveKeys::generate_new(
            "paper_key paper_key",
            path("double_store_and_decrypt_cycle.keys"),
        )
        .unwrap();
        let mut file = open_rwlocked(path("double_store_and_decrypt_cycle.bin")).unwrap();
        let initial_data = b"alksdjflkasjdlfkjasldkfjlaksjdfl";
        let random_data = b"abcdefghijklmnopqrstuvwxyz";
        println!(" >> DBG: {}", random_data.len());

        println!("1: {:?}", keys.get_opening_nonce());
        let mut key = keys.get_sealing_master_key();
        encrypt_file_buffered(&mut file, initial_data, &mut key).expect("unable to encrypt");
        drop(key);

        println!("2: {:?}", keys.get_opening_nonce());
        encrypt_file_buffered(&mut file, random_data, &mut keys.get_sealing_master_key())
            .expect("unable to encrypt");

        let nonce_after_encrypt = keys.get_opening_nonce();
        println!("3: {nonce_after_encrypt:?}");

        let mut key = keys.get_opening_master_key();
        let decrypted = decrypt_file_buffered(&mut file, &mut key);
        drop(key);

        let nonce_after_decrypt = keys.get_opening_nonce();
        println!("4: {nonce_after_decrypt:?}");
        let decrypted = decrypted.expect("unable to decrypt");

        assert_eq!(random_data, &*decrypted);
    }

    #[test]
    fn file_does_not_exist() {
        let vfs = open_dir("bla bla secure key bla", path("file_does_not_exist"))
            .expect("unable to open");
        assert_eq!(Ok(false), vfs.exists("something"));
    }

    #[test]
    fn file_does_exist() {
        let mut vfs =
            open_dir("bla bla secure key bla", path("file_does_exist")).expect("unable to open");
        let _handler = vfs.create("something").unwrap();
        assert_eq!(Ok(true), vfs.exists("something"));
        println!("vfs: {:#?}", vfs);

        drop(vfs);

        let vfs = open_dir("bla bla secure key bla", path("file_does_exist")).unwrap();
        assert_eq!(Ok(true), vfs.exists("something"));
    }

    #[test]
    fn store_string() {
        let clear = String::from("this is a string that I need to store privately");
        let safe_name = "my thing.txt";
        let repo_key = "this is a secure secre key for testing";

        {
            let mut vfs = open_dir(repo_key, path("store_string")).expect("unable to open");

            let mut handler = vfs.create(safe_name).expect("unable to get handler");
            handler
                .write_all(clear.as_bytes())
                .expect("unable to write_all");

            handler.try_flush().unwrap();

            //println!("{vfs:?}");
        }

        {
            println!("\n\n222222222222222222\nsecond opening");
            let mut vfs = open_dir(repo_key, path("store_string")).expect("unable to open");

            let handler = vfs.create(safe_name).unwrap();
            let decrypted = read_to_string(handler).unwrap();

            assert_eq!(clear, decrypted);
            println!("{vfs:?}");
        }
    }

    #[test]
    fn open_two() {
        let mut vfs = open_dir("bla bla secure key bla", path("open_two")).expect("unable to open");

        let mut a = vfs.create("a").unwrap();
        let mut a1 = vfs.create("a1").unwrap();
        let mut b = vfs.create("b").unwrap();
        assert_ne!(a.id, b.id);
        assert_ne!(a.id, a1.id);

        a.set_len(0).unwrap();
        a1.set_len(3).unwrap();
        b.set_len(12).unwrap();

        a.write_all(b"abcdefghijklmnopqrstuvwxyz").unwrap();
        a1.write_all(b"abcdefghijklmnopqrstuvwxyz").unwrap();
        b.write_all(b"01234566789012345678901234").unwrap();
    }

    #[test]
    fn delete_a_file() {
        let mut vfs =
            open_dir("bla bla secure key bla", path("delete_a_file")).expect("unable to open");

        let mut a = vfs.create("a").unwrap();
        let mut b = vfs.create("b").unwrap();
        assert_ne!(a.id, b.id);

        b.set_len(12).unwrap();

        a.write_all(b"abcdefghijklmnopqrstuvwxyz").unwrap();
        b.write_all(b"abcdefghijklmnopqrstuvwxyz").unwrap();

        vfs.flush_all().unwrap();

        let mut b_block_path = path("delete_a_file");
        b_block_path.push("blocks");
        b_block_path.push(&vfs.0.lock().unwrap().metadata.files["b"].real_block_name);
        assert!(b_block_path.exists());

        assert_eq!(Ok(true), vfs.delete("b"));
        assert!(b_block_path.exists() == false);
        assert_eq!(Ok(false), vfs.exists("b"));

        //assert!(false, "{vfs:#?}");
    }

    #[test]
    fn read_many() {
        let mut vfs =
            open_dir("bla bla secure key bla", path("read_many")).expect("unable to open");
        let expected = b"abcdefghijklmnopqrstuvwxyz";

        let mut a = vfs.create("a").unwrap();
        a.set_len(0).unwrap();
        a.write_all(expected).unwrap();
        drop(a);

        println!("\n    Reading First");
        let mut buf = Vec::with_capacity(expected.len());
        let mut a = vfs.create("a").unwrap();
        a.read_to_end(&mut buf).unwrap();
        drop(a);
        assert_eq!(expected, &*buf, "frist read");

        println!("\n    Reading Second");
        buf.clear();
        let mut a = vfs.create("a").unwrap();
        a.read_to_end(&mut buf).unwrap();
        drop(a);
        assert_eq!(expected, &*buf, "second read");
    }
}
