use aead_vfs::{open_dir, CvfsResult};
use core::hint::black_box;
use log::{error, info, warn};
use std::{
    io::{Read, Write},
    process::exit,
};
use structopt::StructOpt;

/// Read some lines of a file
#[derive(Debug, StructOpt)]
struct Cli {
    /// Enable Debug Output
    #[structopt(short, long)]
    debug: bool,
    /// Directory to store
    path: String,
    /// The passphrase for the encryption
    passphrase: String,
    /// Actions: create, read, store, delete, exists
    action: Action,
    /// The encrypted file
    secure_file_path: Option<String>,
    /// content for create and store
    content: Option<String>,
}

#[derive(Debug, Clone, Copy)]
enum Action {
    Create,
    Read,
    Store,
    Delete,
    Exists,
    SpeedTest,
}
impl std::str::FromStr for Action {
    type Err = String;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        use Action::*;
        let action = s.trim().to_lowercase();
        Ok(match &*action {
            "create" => Create,
            "read" => Read,
            "store" => Store,
            "delete" => Delete,
            "exists" => Exists,
            "speedtest" | "speed_test" => SpeedTest,
            unknown => return Err(format!("unknown Action: {unknown:?}")),
        })
    }
}

fn main() {
    let cli = Cli::from_args();

    let level = if cli.debug {
        log::LevelFilter::Trace
    } else {
        log::LevelFilter::Info
    };
    stderrlog::new()
        .verbosity(level)
        .show_level(true)
        .init()
        .unwrap();

    if let Err(e) = run(cli) {
        eprintln!("{e:#?}");
    }
}
fn run(cli: Cli) -> CvfsResult<()> {
    let mut vfs = open_dir(&cli.passphrase, &cli.path)?;

    if cli.debug {
        info!("    ======================");
        info!("{cli:#?}");
        info!("    ======================");
        info!("{vfs:#?}");
        info!("    ======================");
    }

    use Action::*;
    match (cli.action, cli.secure_file_path, cli.content) {
        (Create, None, _)
        | (Read, None, _)
        | (Store, None, _)
        | (Delete, None, _)
        | (Exists, None, _) => {
            error!("Must provide a file argument with action: {:?}", cli.action);
            exit(23);
        }
        (Create, _, None) | (Store, _, None) => {
            error!("Must provide content argument");
            exit(23);
        }
        (Create, Some(secure_file_path), Some(content)) => {
            let mut a = vfs.create(&secure_file_path)?;
            a.set_len(content.len())?;

            a.write_all(content.as_bytes())?;
        }
        (Read, Some(secure_file_path), _) => {
            if vfs.exists(&secure_file_path)? == false {
                eprintln!("secure_file_path {:#?} not found", secure_file_path);
                exit(23);
            }
            let mut fd = vfs.create(&secure_file_path)?;
            let mut buffer = String::new();
            fd.read_to_string(&mut buffer)?;
            info!("{buffer}");
        }
        (Store, Some(secure_file_path), Some(content)) => {
            let mut fd = vfs.create(&secure_file_path)?;
            fd.write_all(content.as_bytes())?;
        }
        (Delete, Some(secure_file_path), _) => {
            let _r = vfs.delete(&secure_file_path)?;
        }
        (Exists, Some(secure_file_path), _) => {
            let r = vfs.exists(&secure_file_path)?;
            if r {
                info!("exists");
            } else {
                warn!("absent");
            }
        }
        (SpeedTest, _, _) => {
            let mb = black_box(128);
            info!("testing with {mb} MB");
            info!("");
            let result = aead_vfs::do_speed_test(mb, &mut vfs)?;
            info!("{result}");
        }
    }

    Ok(())
}
